"""
Utility functions for dictionaries.
"""

def check_dict_structure(
    dict_1: dict,
    dict_2: dict,
    *,
    full_depth: bool = False
) -> bool:
    """
    If ``full_depth`` is ``False`` (default) check that ``dict_2`` contains at least
    the same nested-key structure as ``dict_1`` (extra depth in ``dict_2``
    is ignored).  
    If ``full_depth`` is ``True``, require both dictionaries to match at every
    level.

    Parameters
    ----------
    dict_1
        Template dictionary (sets maximum depth when ``full_depth`` is False).

    dict_2
        Candidate dictionary to compare.

    full_depth
        If ``True``, check that both dictionaries have the same keys and nesting structure at every level.
        If ``False``, check that ``dict_2`` has at least the same keys and nesting structure as ``dict_1``, ignoring extra depth in ``dict_2``.

    Returns
    -------
    ``True`` if both dictionaries have the same keys and nesting structure (ignoring extra depth in ``dict_2`` if ``full_depth`` is ``False``), ``False`` otherwise.
    """

    if not (isinstance(dict_1, dict) and isinstance(dict_2, dict)):
        return False

    def _match(template: dict, candidate: dict) -> bool:
        # Reject extra or missing sibling keys at this depth
        if set(candidate.keys()) != set(template.keys()):
            return False

        for k, v in template.items():
            cand_v = candidate[k]
            if isinstance(v, dict):
                if not isinstance(cand_v, dict):
                    return False
                if not _match(v, cand_v):
                    return False
            else:
                # Template stops here. Extra depth in candidate is OK
                # unless strict checking is requested.
                if full_depth and isinstance(cand_v, dict):
                    return False
        return True

    if full_depth:
        # Two-way comparison
        return _match(dict_1, dict_2) and _match(dict_2, dict_1)

    # One-way comparison
    return _match(dict_1, dict_2)


def all_leaves_are_of_type(
    data,
    target_type
) -> bool:
    """
    Recursively check if all leaves in a nested dictionary or list are of a specific type.

    Parameters
    ----------
    data
        The nested dictionary or list to check.

    target_type
        The type to check for at the leaves.

    Returns
    -------
    ``True`` if all leaves are of the target type, ``False`` otherwise.
    """
    # function only recurses through dicts
    if isinstance(data, dict):
        if not data:  # empty dict fails
            return False
        for value in data.values():
            if not all_leaves_are_of_type(value, target_type):
                return False
        return True

    # once we hit a non-dict, we check the type directly
    return isinstance(data, target_type)



"""
This module annotates gene lists with cell types based on a list of expected cell types. 
Can pass tissue information. 
Can pass cell type information if looking for subtypes with ``subtype=True``.

Based on :func:`~anndict.annotate.cells.de_novo.ai_cell_types_by_comparison`.
"""

from anndata import AnnData

from anndict.utils import enforce_semantic_list
from anndict.llm import retry_call_llm, process_llm_category_mapping
from .base import ai_annotate_by_comparison


def ai_from_expected_cell_types(
    gene_lists: list[list[str]],
    expected_cell_types: list[str] = None,
    cell_types: list[str] | None = None,
    tissues: list[str] = None,
    subtype: bool = False
) -> list[str]:
    """
    Returns cell type labels for multiple lists of marker genes. 
    The output cell type labels are drawn from an input list of expected cell types, ``expected_cell_types``.

    Parameters
    -------------
    gene_lists
        A list containing multiple lists of genes to be described.

    expected_cell_types
        A list of expected cell types to provide context to the LLM. 
        The output labels will draw from this list.

    cell_types
        A list of cell types to provide context to the LLM. 
        Each entry corresponds to a gene list in ``gene_lists``.

    tissues
        A list of tissues of origin to provide context to the LLM. 
        Each entry corresponds to a gene list in ``gene_lists``.

    subtype
        Whether to try to annotate cell type (if ``False``) or subtype (if ``True``).

    Returns
    --------
    The cell type labels generated by an LLM for each gene list.
    """
    if not gene_lists:
        raise ValueError("Empty list of gene lists passed to ai_cell_type_by_comparison")

    if cell_types and not subtype:
        raise ValueError("Cannot provide cell types without subtype=True")

    if not expected_cell_types:
        raise ValueError("Expected cell types must be provided. \
            To generate cell type annotations by comparison without providing \
            expected cell types, use :func:`ai_cell_types_by_comparison`.")

    if not tissues:
        raise ValueError("Tissues must be provided. \
            To generate cell type annotations by comparison without providing \
            expected cell types, use :func:`ai_cell_types_by_comparison`.")

    if len(tissues) != len(gene_lists):
        raise ValueError("len(tissues) must match len(gene_lists).")

    # Enforce semantic_list for each gene list
    for gene_list in gene_lists:
        enforce_semantic_list(gene_list)

    # Prepare the system prompt
    system_prompt = (
        # Description of desired response style
        "You are a terse molecular biologist. You respond in a few words \
        and without restating any part of the question. "

        # Prime with expected cell types
        f"Note that this is {set(tissues)} tissue and we expect to find \
        the following cell types: {'    '.join(expected_cell_types)}."

        # Initial instructions
        f"Compare and contrast gene sets to identify the most likely \
        cell type based on marker genes."
    )

    # Prepare the initial user prompt for contrasting all gene lists
    # initial_prompt = f"Tissue: {tissues}, " if tissue else ""
    # initial_prompt += f"Cell Type: {cell_type}, " if cell_type else ""
    initial_prompt = "Briefly compare and contrast the following gene sets:\n"
    for i, gene_list in enumerate(gene_lists):
        tissue_str = " " + ', '.join(tissues[i]) if tissues and tissues[i] else ""
        cell_type_str = " " + ', '.join(cell_types[i]) if cell_types and cell_types[i] else ""

        initial_prompt += f"{i + 1}){tissue_str}{cell_type_str} {('    '.join(gene_list))}\n"

    # Initialize the conversation
    messages = [
        {"role": "system", "content": system_prompt},
        {"role": "user", "content": initial_prompt}
    ]

    # Get the initial contrast response
    contrast_response = retry_call_llm(
        messages=messages,
        process_response=lambda x: x,
        failure_handler=lambda: "Failed to contrast gene sets",
        call_llm_kwargs={'max_tokens': 300, 'temperature': 0},
        max_attempts=1
    )

    # Append the contrast response to the conversation
    messages.append({"role": "assistant", "content": contrast_response})

    messages.append({"role": "user", "content": "Provide only the new label. "})

    # Process each gene list
    cell_subtype_labels = []
    for i, gene_list in enumerate(gene_lists):
        tissue_str = " " + ', '.join(tissues[i]) if tissues and tissues[i] else ""
        cell_type_str = " " + ', '.join(cell_types[i]) if cell_types and cell_types[i] else ""

        gene_set_prompt = f"What is the cell{tissue_str}{cell_type_str} {'sub' if subtype else ''}type \
            label for the gene set: {('    '.join(gene_list))}?"
        messages.append({"role": "user", "content": gene_set_prompt})

        # Get the subtype label
        subtype_label = retry_call_llm(
            messages=messages,
            process_response=lambda x: x.strip(),
            failure_handler=lambda cell_type_str=cell_type_str, i=i: cell_type_str if cell_types and cell_types[i] else "Unknown",
            call_llm_kwargs={'max_tokens': 50, 'temperature': 0},
            max_attempts=1
        )

        cell_subtype_labels.append(subtype_label)
        messages.append({"role": "assistant", "content": subtype_label})

    # print(f"{messages}")

    # Map the returned labels to the original expected cell types
    ## Construct LLM mapping dictionary from expected_cell_types
    llm_dict = {cell_type: cell_type for cell_type in expected_cell_types}

    ## Use the function to get the mapping of cell_subtype_labels to expected_cell_types
    mapped_labels = process_llm_category_mapping(cell_subtype_labels, llm_dict)

    ## Replace cell_subtype_labels with their mapped values
    updated_cell_subtype_labels = [mapped_labels[label] for label in cell_subtype_labels]

    return updated_cell_subtype_labels

def ai_annotate_from_expected_cell_types(
    adata: AnnData,
    groupby: str,
    n_top_genes: int,
    *,
    expected_cell_types: list[str] = None,
    new_label_column: str ='ai_cell_type_by_comparison',
    cell_type_of_origin_col: str = None,
    tissue_of_origin_col: str = None,
    **kwargs
) -> AnnData:
    """
    Annotate cell types based on the top differentially expressed genes in each cluster.
    The output cell type labels are drawn from a list of expected cell types, ``expected_cell_types``.

    Parameters
    -------------
    adata
        AnnData Annotated data matrix.

    groupby
        Column name in adata.obs for grouping cells.

    n_top_genes
        Number of top genes to consider for annotation.

    expected_cell_types
        A list of expected cell types to provide context to the LLM. 
        The output labels will draw from this list.

    new_label_column
        Name of the column to store the AI-generated cell type labels.

    cell_type_of_origin_col
        Name of a column in ``adata.obs`` that contains the cell type 
        of orgin. Used for context to the LLM.

    tissue_of_origin_col
        Name of a column in ``adata.obs`` that contains the tissue of orgin. 
        Used to provide context to the LLM.

    adt_key
        Used by :func:`adata_dict_fapply` or :func:`adata_dict_fapply_return` 
        when passing this function to them.

    Returns
    -------
    :class:`AnnData` with LLM-generated cell type labels stored in ``adata.obs[label_column]``.

    Examples
    --------

    .. code-block:: python

        import anndict as adt

        # This will annotate each cluster based on its top differentially expressed genes after considering all gene lists in the context of each other.
        # The expected cell types are provided as context, and output will be attempted to be mapped back to those cell types to ensure consistency.
        ai_annotate_from_expected_cell_types(
            adata,
            groupby='leiden',
            n_top_genes=10,
            expected_cell_types=['T cell', 'B cell', 'Macrophage', 'Fibroblast'],  # The list of expected cell types
            new_label_column='ai_cell_type_by_comparison',
            cell_type_of_origin_col='cell_type',
            tissue_of_origin_col='tissue                                           # Tissue information for context. Required.
        )
    """
    if not expected_cell_types:
        raise ValueError("Expected cell types must be provided. \
            To generate cell type annotations by comparison without providing \
            expected cell types, use :func:`ai_annotate_cell_type_by_comparison` instead.")

    if not tissue_of_origin_col:
        raise ValueError("tissue_of_origin_col must be provided. \
        To generate cell type annotations by comparison without providing \
        expected cell types, use :func:`ai_annotate_cell_type_by_comparison` instead.")

    return ai_annotate_by_comparison(func=ai_from_expected_cell_types, adata=adata, groupby=groupby, n_top_genes=n_top_genes, expected_cell_types=expected_cell_types,new_label_column=new_label_column, cell_type_of_origin_col=cell_type_of_origin_col, tissue_of_origin_col=tissue_of_origin_col, **kwargs)



"""
This module contains the function to add a stratification to an existing :class:`AdataDict`.
"""

from __future__ import annotations #allows type hinting without circular dependency

from typing import TYPE_CHECKING

from .adata_dict_utils import to_nested_list, to_nested_tuple
from .adata_dict_fapply import adata_dict_fapply_return
from .build import build_adata_dict

if TYPE_CHECKING:
    from .adata_dict import AdataDict


def add_stratification(
    adata_dict: AdataDict,
    strata_keys: list[str],
    *,
    desired_strata: list | dict | None = None,
) -> AdataDict:
    """
    Split each value of an AnnData dictionary into further subsets based on additional desired strata.

    Parameters
    ------------
    adata_dict
        An :class:`AdataDict`

    strata_keys
        List of column names in `adata.obs` to use for further stratification.

    desired_strata
        List of desired strata values or a dictionary where keys are strata keys and values 
        are lists of desired strata values.

    Returns
    -------
    Nested :class:`AdataDict`, where the top-level is now stratified by ``strata_keys``.

    Raises
    ------
    ValueError
        If any of the `strata_keys` are already in the hierarchy.

    Examples
    --------
    Case 1: Build by Donor first, then add Tissue stratification after

    .. code-block:: python

        import pandas as pd
        import anndict as adt
        from anndata import AnnData
        # Create an example AnnData object
        adata = AnnData(obs=pd.DataFrame({
        >     "Donor": ["Donor1", "Donor1", "Donor2"],
        >     "Tissue": ["Tissue1", "Tissue2", "Tissue1"]
        > }))
        # First, build an AdataDict grouped/stratified by Donor
        strata_keys = ["Donor"]
        adata_dict = adt.build_adata_dict(adata, strata_keys)
        print(adata_dict)
        > {
        >     ("Donor1",): adata_d1,
        >     ("Donor2",): adata_d2,
        > }
        # Then, add a stratification by ``Tissue``
        strata_keys = ["Tissue"]
        adata_dict.add_stratification(strata_keys)
        print(adata_dict)
        > {
        >     ("Tissue1",) : {
        >         ("Donor1",) : adata_d1_t1,
        >         ("Donor2",) : adata_d2_t1
        >         },
        >     ("Tissue2",) : {
        >         ("Donor1",) : adata_d1_t2
        >         }
        > }
        # Note 1 If you wanted a new object instead of modifying the original ``adata_dict``, you can instead do:
        new_adata_dict = adt.add_stratification(adata_dict, strata_keys)

        # Note 2: we can always flatten or rearrange the nesting structure
        adata_dict.set_hierarchy(["Donor","Tissue"])
        print(adata_dict)
        > {
        >     ("Donor1", "Tissue1"): adata_d1_t1,
        >     ("Donor1", "Tissue2"): adata_d1_t2,
        >     ("Donor2", "Tissue1"): adata_d2_t1,
        > }
        # For example, if we want Donor as the top-level index
        adata_dict.set_hierarchy(["Donor",["Tissue"]])
        >             {
        >     ("Donor1",) : {
        >         ("Tissue1",) : adata_d1_t1,
        >         ("Tissue2",) : adata_d1_t2
        >         },
        >     ("Donor2",) : {
        >         ("Tissue1",) : adata_d2_t1
        >         }
        > }
    """

    # Get the hierarchy and check for redundant stratification
    cached_hierarchy = adata_dict.hierarchy

    for strata_key in strata_keys:
        if strata_key in adata_dict.flatten_nesting_list(to_nested_list(cached_hierarchy)):
            raise ValueError(f"adata_dict is already stratified by {strata_key}")

    # Flatten the adata_dict
    adata_dict.flatten()
    flat_hierarchy = to_nested_list(adata_dict.hierarchy)


    # Split the adata_dict
    adata_dict = adata_dict_fapply_return(adata_dict, build_adata_dict,
                                         strata_keys=strata_keys,
                                         desired_strata=desired_strata,
                                         return_as_adata_dict=True)

    # directly set the hierarchy of the updated adata_dict
    adata_dict._hierarchy = to_nested_tuple([*flat_hierarchy, [strata_keys]]) # pylint: disable=protected-access

    # Create new hierarchy with strata_keys at top level
    new_hierarchy = [*strata_keys, to_nested_list(cached_hierarchy)]

    # Set the new hierarchy
    adata_dict.set_hierarchy(new_hierarchy)

    return adata_dict



"""
Functions to plot the training history of stabilizing classifier, 
and a bar graph of the number of label differences between two columns.
"""

import matplotlib.pyplot as plt

from matplotlib.figure import Figure, Axes
from anndata import AnnData

def plot_training_history(
    history: list[float],
    adt_key: tuple[str,...] | None = None
) -> tuple[Figure, Axes]:
    """
    Plot the training history of a model, showing percent label change versus iteration.

    Parameters
    ------------
    history
        The training history. From :func:`anndict.utils.stabilizing_classifier`.

    adt_key
        Used by :func:`adata_dict_fapply` when wrapping this function.

    Returns
    ---------
    The figure.

    See Also
    ---------
    :func:`anndict.utils.stabilizing_classifier` : The function that returns ``history``.
    """
    fig, ax = plt.subplots(figsize=(10, 6))
    ax.plot(history, marker='o')
    ax.set_title(f'Percent Label Change vs. Iteration - {adt_key}')
    ax.set_xlabel('Iteration')
    ax.set_ylabel('Percent Label Change')
    ax.grid(True)
    return fig, ax


def plot_label_changes(
    adata: AnnData,
    true_label_key: str,
    predicted_label_key: str,
    plot_percentage: bool = True,
    stratum: str = None
) -> tuple[Figure, Axes]:
    """
    Plot the changes between true and predicted labels in an AnnData object.

    Parameters
    ------------
    adata
        An :class:`AnnData`.

    true_label_key
        Key for the true labels in ``adata.obs``.

    predicted_label_key
        Key for the predicted labels in ``adata.obs``.

    plot_percentage
        If ``True``, plot the percentage of labels changed. If ``False``, plot the count of labels changed.

    stratum
         Title for the plot, often used to indicate the stratum. Default is None.

    Returns
    -------
    The figure and axis of the plot.
    """
    # Extract the series from the AnnData object's DataFrame
    data = adata.obs[[predicted_label_key, true_label_key]].copy()

    # Convert to categorical with a common category set
    common_categories = list(set(data[true_label_key].cat.categories).union(set(data[predicted_label_key].cat.categories)))
    data[true_label_key] = data[true_label_key].cat.set_categories(common_categories)
    data[predicted_label_key] = data[predicted_label_key].cat.set_categories(common_categories)

    # Add a mismatch column that checks whether the predicted and true labels are different
    data['Changed'] = data[true_label_key] != data[predicted_label_key]

    # Group by true label key and calculate the sum of mismatches or the mean if plot_percentage
    if plot_percentage:
        change_summary = data.groupby(true_label_key, observed=False)['Changed'].mean()
    else:
        change_summary = data.groupby(true_label_key, observed=False)['Changed'].sum()

    # Sort the summary in descending order
    change_summary = change_summary.sort_values(ascending=False)

    # Plotting
    fig, ax = plt.subplots(figsize=(10, 6))
    change_summary.plot(kind='bar', color='red', ax=ax)
    ax.set_xlabel(true_label_key)
    ax.set_ylabel('Percentage of Labels Changed' if plot_percentage else 'Count of Labels Changed')
    ax.set_title(stratum)
    ax.set_xticklabels(change_summary.index, rotation=90)
    plt.xticks(fontsize=8)
    plt.show()

    return fig, ax



"""
Anndata-specific ``Predicoder`` that remembers which ``adata.obsm`` key it was trained on.
"""
# pylint: disable=arguments-renamed
#disable pylint false positives

import numpy as np

from sklearn.base import ClassifierMixin
from sklearn.preprocessing import LabelEncoder
from anndata import AnnData

from .predicoder import Predicoder

class AdataPredicoder(Predicoder):
    """
    A layer-aware Predicoder for AnnData objects that stores predictions directly in adata.obs.
    
    Parameters
    ----------
    classifier
        The classifier to wrap
    label_encoder
        The label encoder used to transform labels
    feature_key
        Key of data to use in ``adata.obsm``, or ``'use_X'`` to use ``adata.X``
    
    See Also
    --------
    :class:`Predicoder` : the base class.
    """

    def __init__(
        self,
        classifier: ClassifierMixin,
        label_encoder: LabelEncoder,
        feature_key: str,
    ):
        super().__init__(classifier, label_encoder)
        self.feature_key = feature_key

    def _get_x(self, adata: AnnData) -> np.ndarray:
        """Extract features from the AnnData object."""
        if self.feature_key == "use_X":
            return adata.X
        if self.feature_key not in adata.obsm:
            raise KeyError(
                f"Feature key '{self.feature_key}' not found in adata.obsm. "
                f"Available keys: {list(adata.obsm.keys())}"
            )
        return adata.obsm[self.feature_key]

    def predict(
        self,
        adata: AnnData,
        new_column_name: str = 'predicted_label'
    ) -> None:
        """
        Predict labels and store them in ``adata.obs[new_column_name]``.
        
        Parameters
        ----------
        adata
            An :class:`AnnData`.

        new_column_name
            Key to store predictions in ``adata.obs``

        Returns
        -------
        None

        Notes
        -----
        Updates ``adata`` in-place with predictions, stored in ``adata.obs[new_column_name]``.
        """
        x = self._get_x(adata)
        predictions = super().predict(x)
        adata.obs[new_column_name] = predictions

    def predict_proba(
        self,
        adata: AnnData,
        new_column_name: str = 'predicted_prob'
    ) -> None:
        """
        Get predicted probabilities of membership for each class and store them in ``adata.obs``.
        
        Parameters
        ----------
        adata
            An :class:`AnnData`.

        new_column_name
            Prefix for probability columns in ``adata.obs`` (default: 'predicted_prob')
            Each class probability will be stored as '{new_column_name}_{class_name}'
            
        Returns
        -------
        None

        Notes
        -----
        Updates ``adata`` in-place with predicted probabilities, stored in ``adata.obs``.
        """
        x = self._get_x(adata)
        probabilities = super().predict_proba(x)

        # Store probability for each class
        for idx, class_name in enumerate(self.label_encoder.classes_):
            col_name = f"{new_column_name}_{class_name}"
            adata.obs[col_name] = probabilities[:, idx]



"""
This module handles label transfer via sklearn models.
"""

from typing import Type, Literal

from sklearn.base import ClassifierMixin
from anndata import AnnData

from anndict.utils.stabilizing_classifier import stable_label_adata

from .adata_predicoder import AdataPredicoder


def train_label_classifier(
    adata: AnnData,
    label_key: str,
    feature_key: str | Literal["use_X"],
    classifier_class: Type[ClassifierMixin],
    *,
    random_state: int | None = None,
    **kwargs
) -> AdataPredicoder:
    """
    Trains a classifier on the given ``adata``; used internally by :func:`transfer_labels_using_classifier`.

    Parameters
    ----------
    adata
        An :class:`AnnData` containing the original labels.
        A classifier will be trained on this ``adata``.

    label_key
        Key in ``adata.obs`` containing the original labels.

    feature_key
        Key of data to use in ``adata.obsm``, or ``'use_X'`` to use ``adata.X``.

    classifier_class
        Any classifier inheriting from :class:`sklearn.base.ClassifierMixin`. 
        Pass as a class, e.g. ``LogisticRegression``, and not an already-instantiated object.

    random_state
        random state seed passed to :func:`stable_label_adata`.

    **kwargs
        Additional keyword arguments passed to the classifier constructor.

    Returns
    -------
    A :class:`AdataPredicoder`, containing the trained classifier and label encoder/decoder.

    See Also
    --------
    :class:`AdataPredicoder` : The container class for classifier+label encoder/decoder.
    :func:`stable_label_adata` : The function that trains the classifier on ``adata``.

    Examples
    --------
    **Case 1: Using a logistic regression classifier**

    .. code-block:: python

        import anndict as adt
        from sklearn.linear_model import LogisticRegression

        train_label_classifier(
            adata=adata,
            label_key='cell_type',
            feature_key='X_pca',
            classifier_class=LogisticRegression,
            penalty='l2', #one kwarg for LogisticRegression
            fit_intercept=True, #another kwarg for LogisticRegression
        )

    **Case 2: Using a random forest classifier**

    .. code-block:: python

        import anndict as adt
        from sklearn.ensemble import RandomForestClassifier

        train_label_classifier(
            adata=adata,
            label_key='cell_type',
            feature_key='X_pca',
            classifier_class=RandomForestClassifier,
            n_estimators=1000, #one kwarg for RandomForestClassifier
            max_features='sqrt', #another kwarg for RandomForestClassifier
        )

    """
    # create a classifier for this stratum
    classifier = classifier_class(random_state=random_state, **kwargs)

    classifier, _, _, _, label_encoder = stable_label_adata(
        adata, label_key, feature_key, classifier, max_iterations=1, random_state=random_state
    ) # max_iterations=1 causes classifier to only train one round, without label resampling

    predicoder = AdataPredicoder(classifier, label_encoder, feature_key)

    return predicoder

def transfer_labels_using_classifier(
    origin_adata: AnnData,
    destination_adata: AnnData,
    origin_label_key: str,
    feature_key: str | Literal["use_X"],
    classifier_class: Type[ClassifierMixin],
    new_column_name: str = 'predicted_label',
    random_state: int | None = None,
    **kwargs
) -> AdataPredicoder:
    """
    Transfers labels from ``origin_adata`` to ``destination_adata`` 
    using a classifier of type ``classifier_class``.

    Supported classifiers include any ``sklearn`` classifier inheriting 
    from :class:`sklearn.base.ClassifierMixin`.

    Parameters
    ----------
    origin_adata
        An :class:`AnnData` containing the original labels.
        A classifier will be trained on this adata.

    destination_adata
        An :class:`AnnData` containing the new cells to be 
        labeled. Must contain the same ``.obsm[feature_key]`` 
        as ``origin_adata`` if ``feature_key`` is not ``'use_X'``.

    origin_label_key
        Key in ``origin_adata.obs`` containing the original labels.

    feature_key
        Key of data to use in ``origin_adata.obsm``, or ``'use_X'`` to use ``origin_adata.X``.

    classifier_class
        Any classifier inheriting from :class:`sklearn.base.ClassifierMixin`. 
        Pass as a class, e.g. ``LogisticRegression``, and not an already-instantiated object.

    new_column_name
        The name of the new column in ``destination_adata.obs`` where 
        the predicted labels will be stored.
    
    random_state
        random state seed passed to :func:`stable_label_adata`.
    
    **kwargs
        Additional keyword arguments passed to the classifier constructor.

    Returns
    -------
    A :class:`AdataPredicoder` that contains the 
    trained classifier, and automatically decodes 
    predicted labels into text labels. Can be used 
    to calculate class membership probabilities or 
    predict on other :class:`AnnData`.

    Notes
    -----
    Modifies ``destination_adata`` in-place.

    See Also
    --------
    :class:`AdataPredicoder` : The container class for classifier+label 
        encoder/decoder.
    :func:`train_label_classifier` : The function that trains the 
        classifier on ``origin_adata``.

    Examples
    --------
    **Case 1: Using a logistic regression classifier**

    .. code-block:: python

        import anndict as adt
        from sklearn.linear_model import LogisticRegression

        transfer_labels(
            origin_adata=origin_adata,
            destination_adata=destination_adata,
            origin_label_key='cell_type',
            feature_key='X_pca',
            classifier_class=LogisticRegression,
            new_column_name='predicted_label',
            penalty='l2', #one kwarg for LogisticRegression
            fit_intercept=True, #another kwarg for LogisticRegression
        )

    **Case 2: Using a random forest classifier**

    .. code-block:: python

        import anndict as adt
        from sklearn.ensemble import RandomForestClassifier

        transfer_labels(
            origin_adata=origin_adata,
            destination_adata=destination_adata,
            origin_label_key='cell_type',
            feature_key='X_pca',
            classifier_class=RandomForestClassifier,
            new_column_name='predicted_label',
            n_estimators=1000, #one kwarg for RandomForestClassifier
            max_features='sqrt', #another kwarg for RandomForestClassifier
        )

    """
    adata_predicoder = train_label_classifier(
        adata=origin_adata, feature_key=feature_key, label_key=origin_label_key, classifier_class=classifier_class, random_state=random_state, **kwargs)

    adata_predicoder.predict(destination_adata, new_column_name=new_column_name)
    return adata_predicoder



"""
This module contains utility functions for label comparison.
"""

import pandas as pd

from anndata import AnnData

def create_label_df(
    adata: AnnData,
    cols1: list[str],
    cols2: list[str]
) -> pd.DataFrame:
    """
    Creates a :class:`DataFrame` of unique label combinations from 
    the specified columns in ``cols1`` and ``cols2``, only including 
    combinations that exist in ``adata.obs[[*cols1, *cols2]]``.

    Parameters
    ----------
    adata
        An :class:`AnnData`.

    cols1
        List of columns to compare with cols2.

    cols2
        List of columns to compare with cols1.

    Returns
    -------
    :class:`DataFrame` containing unique combinations of the specified columns.
    """
    # Combine all columns
    all_cols = cols1 + cols2

    # Get unique combinations that exist in adata.obs
    unique_combinations = adata.obs[all_cols].drop_duplicates()

    # Melt the DataFrame to get all combinations in two columns
    melted_df = pd.melt(unique_combinations,
                        id_vars=cols1,
                        value_vars=cols2,
                        var_name='col2_name',
                        value_name='col2')

    # Melt again to get col1 in a single column
    result_df = pd.melt(melted_df,
                        id_vars=['col2_name', 'col2'],
                        value_vars=cols1,
                        var_name='col1_name',
                        value_name='col1')

    # Keep only the relevant columns and drop duplicates
    result_df = result_df[['col1', 'col2']].drop_duplicates()

    return result_df
